import os
import json
from itertools import chain
from dektools.file import write_file, read_file
from dektools.startup.cache import StartupCache
from ...const import weight_dict, font_props_default
from ...type import get_afm_meta, get_ttf_meta
from ...path import iter_all_fonts, iter_dir_fonts


class Manager(StartupCache):
    __version__ = 'v0.1.0'
    done_name = __version__

    def __init__(self, paths=None):
        self.paths = paths if paths else []
        if self._is_done():
            self.font_list = json.loads(read_file(self.path_data))
        else:
            self.font_list = []
        super().__init__()

    @property
    def path_data(self):
        return self.path_index / self.__version__ / 'fonts.json'

    def _cache_handler(self):
        result = []
        for file in chain(iter(iter_all_fonts()), iter(iter_dir_fonts(*self.paths))):
            if os.path.splitext(file)[-1].lower() == ".afm":
                info = get_afm_meta(file)
            else:
                info = get_ttf_meta(file)
            result.append({'path': file, 'props': info})
        self.font_list = result
        write_file(self.path_data, s=json.dumps(self.font_list))

    def find_family(self, family, **kwargs):
        name_list = [name.strip() for name in family.split(',')]
        entries = [{**font_props_default, 'name': name, **kwargs} for name in name_list]
        return self.find(entries)

    def find(self, entries):
        for props in entries:
            props['weight'] = weight_dict.get(props['weight'], props['weight'])
        for item in self.font_list:
            if item['props'] in entries:
                return item['path']



import re
from .manager.default import font_manager


def is_font_family(s):
    return re.fullmatch(r'^[ 0-9a-zA-Z_-]+$', s)


def get_font_source(name):
    if name is None:
        name = 'SimSun'
    if isinstance(name, str):
        if is_font_family(name):
            with font_manager:
                font_path = font_manager.find_family(name)
            if font_path is None:
                raise FileNotFoundError(f"Can't find font path by name: {name}")
        else:
            font_path = name
        with open(font_path, 'rb') as f:
            return f.read()
    elif isinstance(name, bytes):
        return name
    else:
        return name.getvalue()



weight_dict = {
    'ultralight': 100,
    'light': 200,
    'normal': 400,
    'regular': 400,
    'book': 400,
    'medium': 500,
    'roman': 500,
    'semibold': 600,
    'demibold': 600,
    'demi': 600,
    'bold': 700,
    'heavy': 800,
    'extra bold': 800,
    'black': 900,
}

weight_set = set(weight_dict)

font_props_default = {
    'style': 'normal',
    'weight': weight_dict['normal'],
    "size": "scalable",
    "stretch": "normal",
    "variant": "normal",
}



from ...const import weight_dict, font_props_default
from . import _afm


def get_afm_meta(font_path):
    with open(font_path, "rb") as fh:
        font = _afm.AFM(fh)
    name = font.get_familyname()
    fontname = font.get_fontname().lower()

    #  Styles are: italic, oblique, and normal (default)

    if font.get_angle() != 0 or 'italic' in name.lower():
        style = 'italic'
    elif 'oblique' in name.lower():
        style = 'oblique'
    else:
        style = font_props_default['style']

    #  Variants are: small-caps and normal (default)

    # !!!!  Untested
    if name.lower() in ['capitals', 'small-caps']:
        variant = 'small-caps'
    else:
        variant = font_props_default['variant']

    weight = font.get_weight().lower()
    if weight not in weight_dict:
        weight = font_props_default['weight']

    #  Stretch can be absolute and relative
    #  Absolute stretches are: ultra-condensed, extra-condensed, condensed,
    #    semi-condensed, normal, semi-expanded, expanded, extra-expanded,
    #    and ultra-expanded.
    #  Relative stretches are: wider, narrower
    #  Child value is: inherit
    if 'demi cond' in fontname:
        stretch = 'semi-condensed'
    elif any(word in fontname for word in ['narrow', 'cond']):
        stretch = 'condensed'
    elif any(word in fontname for word in ['wide', 'expanded', 'extended']):
        stretch = 'expanded'
    else:
        stretch = font_props_default['stretch']

    #  Sizes can be absolute and relative.
    #  Absolute sizes are: xx-small, x-small, small, medium, large, x-large,
    #    and xx-large.
    #  Relative sizes are: larger, smaller
    #  Length value is an absolute font size, e.g., 12pt
    #  Percentage values are in 'em's.  Most robust specification.

    #  All AFM fonts are apparently scalable.

    return {
        'name': name,
        'style': style,
        'weight': weight_dict.get(weight, weight),
        "size": font_props_default['size'],
        "stretch": stretch,
        "variant": variant,
    }



import os
import time
import subprocess
from pathlib import Path
from just_playback.playback import Playback as PlaybackBase, MiniaudioError


class Playback(PlaybackBase):
    def play_sync(self):
        self.play()
        self.waiting()

    def waiting(self):
        while self.playing:
            time.sleep(0)


def play_file(filepath, ignore=True):
    if ignore:
        if not is_device_supported():
            return
    try:
        playback = Playback()
    except MiniaudioError:
        if ignore:
            return
        else:
            raise
    playback.load_file(filepath)
    playback.play_sync()


path_res = Path(__file__).resolve().parent / 'res'


def play_res(name, path=None, ignore=True):
    path = path or path_res
    for file in os.listdir(path):
        if os.path.splitext(file)[0] == name:
            play_file(os.path.join(path, file), ignore)


def is_device_supported():
    if os.name != 'nt':
        result = subprocess.getoutput('lspci')
        desc_list = [x.rsplit(':', 1)[0].split(' ', 1)[-1].strip() for x in result.split('\n')]
        return 'Audio device' in desc_list
    return True



import string
from typing import Iterable, Literal
from io import BytesIO
from PIL import Image as PilImage, ImageDraw, ImageFont
from dektools.func import FuncAnyArgs
from ...font import get_font_source


def get_spacing(font, spacing=None):
    if spacing is None:
        spacing = 1.0
    if isinstance(spacing, float):
        spacing = font.font.descent * spacing
    return spacing


def get_font(src, size):
    return ImageFont.truetype(BytesIO(get_font_source(src)), size)


def generate_image(text: str, font, angle=0, color=None, bgc=None, spacing=None):
    color = color or (0, 0, 0)
    bgc = bgc or (255, 255, 255)
    spacing = get_spacing(font, spacing)
    image = PilImage.new('RGB', (1, 1))
    draw = ImageDraw.Draw(image)
    box = draw.multiline_textbbox((0, 0), text, font=font, spacing=spacing)
    image = PilImage.new('RGBA' if len(bgc) == 4 else 'RGB', (box[2], int(box[3] + font.font.descent)), bgc)
    draw = ImageDraw.Draw(image)
    draw.text(box[:2], text, font=font, fill=color, spacing=spacing)
    if angle:
        image = image.rotate(angle, expand=True, fillcolor=bgc)
    return image


def get_text_size(font, text):
    bbox = font.getbbox(text)
    return bbox[2] - bbox[0], bbox[3] - bbox[1]


whitespace = set(string.whitespace)


def get_text_box_info(font, texts: Iterable[str], width):
    def append(_w, _h, _word):
        check_the_last()
        widths.append(_w)
        heights.append(_h)
        lines.append(_word)

    def check_the_last():
        if heights and heights[-1] == 0:
            heights[-1] = jh

    lines = ['']
    widths = [0]
    heights = [0]
    jw, jh = get_text_size(font, 'j')
    for word in texts:
        w, h = get_text_size(font, lines[-1] + word)
        if word == '\n':
            append(w, h, word)
        elif w > width + jw / 2:
            append(w, h, word)
        else:
            if lines[-1] == '\n':
                append(0, 0, '')
            lines[-1] += word
            widths[-1] = w
            heights[-1] = h
    check_the_last()
    return lines, widths, heights


def get_text_box_height(font, texts: Iterable[str], width, spacing=None):
    spacing = get_spacing(font, spacing)
    lines, _, heights = get_text_box_info(font, texts, width)
    cursor = 0
    for i in range(len(lines)):
        cursor += heights[i] + spacing
    return int(cursor)


def text_in_box(draw: ImageDraw, texts: Iterable[str], xy, width, font, spacing=None, color=None,
                align: Literal['left', 'right', 'middle'] = 'left', hidden=None):
    if color is None:
        color = (0, 0, 0)
    spacing = get_spacing(font, spacing)
    lines, widths, heights = get_text_box_info(font, texts, width)
    x, y = xy
    cursor = 0
    hidden = FuncAnyArgs(hidden) if hidden else None
    for i, line in enumerate(lines):
        if align == 'left':
            offset = 0
        elif align == 'right':
            offset = width - widths[i]
        else:
            offset = (width - widths[i]) // 2
        if not hidden or not hidden(cursor + heights[i], cursor, i, len(lines)):
            draw.text((x + offset, y + cursor), line, font=font, fill=color)
        cursor += heights[i] + spacing


class Text:
    def __init__(self, s: str):
        self.content = s

    @property
    def block(self):
        return list(self.content)

    @property
    def words(self):
        result = []
        last_ascii = False
        for c in self.content:
            if c.isascii():
                if c.isalpha():
                    if last_ascii:
                        result[-1] += c
                    else:
                        result.append(c)
                    last_ascii = True
                else:
                    result.append(c)
                    last_ascii = False
            else:
                result.append(c)
                last_ascii = False
        return result



from contextlib import redirect_stderr
from fontTools import ttLib
from ..const import weight_dict, weight_set, font_props_default


def get_ttf_meta(font_path):
    font = ttLib.TTFont(font_path, ignoreDecompileErrors=True, fontNumber=0)
    with redirect_stderr(None):
        names = font['name'].names
    details = {}
    for x in names:
        if x.langID == 0 or x.langID == 1033:
            try:
                details[x.nameID] = x.toUnicode()
            except UnicodeDecodeError:
                details[x.nameID] = x.string.decode(errors='ignore')
    sw = set(x.lower() for x in details[2].split())
    ws = weight_set & sw
    if ws:
        weight = next(iter(ws))
    else:
        weight = font_props_default['weight']
    ss = sw - ws
    if ss:
        style = next(iter(ss))
    else:
        style = font_props_default['style']
    return {
        **font_props_default,
        'name': details[1],
        'style': style,
        'weight': weight_dict.get(weight, weight),
    }



from dektools.typer import command_version
from . import app
from . import audio as audio_command
from . import image as image_command
from . import font as font_command

command_version(app, __name__)

app.add_typer(audio_command.app, name='audio')
app.add_typer(image_command.app, name='image')
app.add_typer(font_command.app, name='font')


def main():
    app()



import os
import typer
from typing import List
from typing_extensions import Annotated
from dektools.file import read_file, normal_path

app = typer.Typer(add_completion=False)

@app.command()
def trans(src, outputs: Annotated[List[str], typer.Argument()] = None, sizes=''):
    if not outputs:
        return
    if os.path.splitext(src)[-1].lower() == '.svg':
        from ..image.svg import load_svg
        image = load_svg(read_file(src))
    else:
        from PIL import Image
        image = Image.open(src)
    from ..image.core import resize_image
    resize_image(image, outputs, parse_sizes(sizes))


def parse_sizes(sizes):
    if sizes:
        ss = [int(n) for n in sizes.split(',')]
        return list(zip(ss[::2], ss[1::2]))


@app.command()
def psd(src, dest: Annotated[str, typer.Argument()] = ""):
    from ..image.psd import PsdCanvas
    src = normal_path(src)
    dest = dest if dest else os.path.join(src + '.dump')
    pc = PsdCanvas.load(src)
    pc.dump(dest)



import typer

app = typer.Typer(add_completion=False)


@app.command()
def play(path):
    from ..audio.core import play_file
    play_file(path)


@app.command()
def notify(name, path=None):
    from ..audio.core import play_res
    play_res(name, path)



import os
from pathlib import Path
from find_system_fonts_filename import get_system_fonts_filename, FindSystemFontsFilenameException
from dektools.ext.font import font_extensions

path_res = Path(__file__).resolve().parent.parent / 'res'


def iter_dir_fonts(*paths):
    for path in paths:
        if os.path.isdir(path):
            for root, _, filenames in os.walk(path):
                for name in filenames:
                    if os.path.splitext(name)[-1].lower() in font_extensions:
                        yield os.path.join(root, name)


def iter_sys_fonts():
    try:
        yield from get_system_fonts_filename()
    except FindSystemFontsFilenameException:
        pass


def iter_res_fonts():
    yield from iter_dir_fonts(path_res)


def iter_all_fonts():
    yield from iter_res_fonts()
    yield from iter_sys_fonts()



import fitz
from copy import deepcopy
from PIL import Image
from lxml import etree
from svglib.svglib import SvgRenderer
from reportlab.graphics import renderPDF
from svglib.fonts import FontMap
from ...font.const import font_props_default
from ...font.manager.default import font_manager


def get_svg_meta(root, fonts=None):
    def walk(node):
        ff = node.attrib.get('font-family')
        if ff and ff not in fonts:
            kwargs = {}
            for attr in ('style', 'weight', 'stretch', 'variant'):
                kwargs[attr] = node.attrib.get(f'font-{attr}') or font_props_default[attr]
            with font_manager:
                font_path = font_manager.find_family(ff, **kwargs)
            fonts[ff] = dict(
                font_path=font_path,
                weight=kwargs['weight'],
                style=kwargs['style'])
        for child in node.getchildren():
            walk(child)

    _, _, w, h = root.attrib['viewBox'].split()
    if fonts:
        fonts = deepcopy(fonts)
    else:
        fonts = {}
    walk(root)
    return (int(w), int(h)), fonts


# https://github.com/lovell/sharp/issues/1996#issuecomment-561656817
# No support `dominant-baseline``alignment-baseline`, use `dy=".25em"` to shift the <text/> along the y-axis
def load_svg(content: str | bytes, width=None, height=None, fonts=None):  # Remove <?xml ...> when `content` is a string
    parser = etree.XMLParser(resolve_entities=False, strip_cdata=False, recover=True)
    svg_root = etree.XML(content, parser=parser)
    size, fonts = get_svg_meta(svg_root, fonts)
    dpi = None
    k = 0.013889
    if width is not None:
        dpi = width / (size[0] * k)
    if height is not None and dpi is not None:
        dpi = height / (size[1] * k)
    if dpi is not None:
        dpi = round(dpi)
    font_map = FontMap()
    for ff, kwargs in fonts.items():
        font_map.register_font(ff, **kwargs)
    drawing = SvgRenderer('', font_map=font_map).render(svg_root)
    pdf = renderPDF.drawToString(drawing)
    doc = fitz.Document(stream=pdf)
    pix = doc.load_page(0).get_pixmap(alpha=True, dpi=dpi)
    cspace = pix.colorspace
    if cspace is None:
        mode = "L"
    elif cspace.n == 1:
        mode = "L" if pix.alpha == 0 else "LA"
    elif cspace.n == 3:
        mode = "RGB" if pix.alpha == 0 else "RGBA"
    else:
        mode = "CMYK"
    return Image.frombytes(mode, (pix.width, pix.height), pix.samples)



import qrcode
from qrcode.image.styledpil import StyledPilImage
from dektools.dict import assign

default_props = {
    'version': 1,
    'error': qrcode.constants.ERROR_CORRECT_L,
    'box': 10,
    'border': 0,
    'color': 'black',
    'bgc': 'white',
    'embed': None,
    'fit': True
}


def make_qr(**kwargs):
    props = assign(default_props, kwargs)
    qr = qrcode.QRCode(
        version=props['version'],
        error_correction=props['error'],
        box_size=props['box'],
        border=props['border'],
    )
    qr.add_data(props['text'])
    qr.make(fit=props['fit'])
    return qr.make_image(
        fill_color=props['color'],
        back_color=props['bgc'],
        image_factory=StyledPilImage,
        embeded_image=props['embed'] or None
    )



import os
import re
import json
from dataclasses import dataclass
from collections import OrderedDict
from PIL import Image, ImageDraw
from psd_tools import PSDImage
from psd_tools.api.layers import Layer, TypeLayer
from dektools.file import sure_dir, write_file
from dektools.str import slugify, replace, Unicode
from dektools.dict import dict_merge, assign
from ...font.const import font_props_default
from ...font.manager.default import font_manager
from ..text import text_in_box, Text, get_font, generate_image, get_text_box_height


@dataclass
class Rect:
    left: int
    top: int
    width: int
    height: int

    @property
    def right(self):
        return self.left + self.width

    @property
    def bottom(self):
        return self.top + self.height


class Node:
    default_options = {}

    def __init__(self, rect: Rect, content, options=None):
        self.rect = rect
        self.content = content
        self.options = assign(self.default_options, {} if options is None else options)

    def update(self, content):
        self.content = content

    def update_options(self, options):
        dict_merge(self.options, options)

    def render(self, canvas: 'PsdCanvas', draw: ImageDraw, image: Image):
        raise NotImplementedError()

    def dump(self, canvas: 'PsdCanvas', name, path):
        raise NotImplementedError()

    @staticmethod
    def _save(image: Image, path, name):
        sure_dir(path)
        image.save(os.path.join(path, name + '.png'))


class ImageNode(Node):
    def update(self, content):
        if isinstance(content, str):
            from ..qr import make_qr
            content = make_qr(text=content)
        self.content = content.resize(self.content.size, Image.Resampling.LANCZOS)

    def render(self, canvas: 'PsdCanvas', draw: ImageDraw, image: Image):
        image.paste(self.content, (self.rect.left, self.rect.top), self.content.convert('RGBA'))

    def dump(self, canvas: 'PsdCanvas', name, path):
        self._save(self.content, path, name)


class TextNode(Node):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.args = self._get_args()

    def _get_args(self):
        data = self.content[0]
        text, font, color = data['text'], data['font'], data['color']
        wrap = re.search(r'\S+[\r\n]+\S+', text)
        fp = font_manager.find_family(font['name'], weight=font['weight'], style=font['style'])
        font = get_font(fp, font['size'])
        return wrap, font, color

    @property
    def text(self):
        return replace(self.content[0]['text'], {'\r': '', '\n': ''})

    def update(self, content):
        self.content[0]['text'] = content

    def render(self, canvas: 'PsdCanvas', draw: ImageDraw, image: Image):
        text, (wrap, font, color) = self.text, self.args
        if wrap:
            text_in_box(draw, Text(text).block, (self.rect.left, self.rect.top), self.rect.width, font=font,
                        color=color)
        else:
            draw.text((self.rect.left, self.rect.top), text=text, fill=color, font=font)

    def dump(self, canvas: 'PsdCanvas', name, path):
        text, (wrap, font, color) = self.text, self.args
        if wrap:
            texts = Text(text).block
            height = get_text_box_height(font, texts, self.rect.width)
            image = Image.new('RGBA', (self.rect.width, height), (0, 0, 0, 0))
            draw = ImageDraw.Draw(image)
            text_in_box(draw, texts, (0, 0), self.rect.width, font=font, color=color)
        else:
            image = generate_image(text, font=font, color=color, bgc=(0, 0, 0, 0))
        self._save(image, path, name)
        write_file(os.path.join(path, name + '.txt'), s=text)


class PsdCanvas:
    image_node_cls = ImageNode
    text_node_cls = TextNode

    def __init__(self, width, height, nodes: OrderedDict):
        self.width = width
        self.height = height
        self.nodes = nodes

    def __str__(self):
        return self.__class__.__name__ + str(json.dumps(
            {k: f"<{v.__class__.__name__}>" for k, v in self.nodes.items()},
            indent=2, ensure_ascii=False))

    def update(self, data):
        for name, content in data.items():
            self.nodes[name].update(content)

    def update_options(self, data):
        for name, options in data.items():
            self.nodes[name].update_options(options)

    def render(self):
        image = Image.new('RGBA', (self.width, self.height), (0, 0, 0, 0))
        draw = ImageDraw.Draw(image)
        for node in self.nodes.values():
            node.render(self, draw, image)
        return image

    def dump(self, path):
        for name, node in self.nodes.items():
            node.dump(self, name, path)
        path = os.path.join(path, 'meta', 'index.png')
        sure_dir(os.path.dirname(path))
        self.render().save(path)

    @classmethod
    def load(cls, file, short: int | None = 15, suffix: str = '-'):
        psd = PSDImage.open(file)
        names = get_psd_names(psd, short, suffix)
        nodes = OrderedDict()
        for layer in psd:
            layer: Layer
            name = names[layer.layer_id]
            rect = Rect(layer.left, layer.top, layer.width, layer.height)
            if isinstance(layer, TypeLayer):
                yy = layer.transform[3]  # https://github.com/psd-tools/psd-tools/issues/176#issuecomment-718514769
                data_list = []
                text = layer.engine_dict['Editor']['Text'].value
                fs = layer.resource_dict['FontSet']
                run_length = layer.engine_dict['StyleRun']['RunLengthArray']
                run_data = layer.engine_dict['StyleRun']['RunArray']
                index = 0
                for length, style in zip(run_length, run_data):
                    sub_text = text[index:index + length]
                    stylesheet = style['StyleSheet']['StyleSheetData']
                    color = get_font_color(stylesheet)
                    font = dict(
                        name=fs[stylesheet['Font']]['Name'].value,
                        size=stylesheet['FontSize'] * yy,
                        weight='bold' if stylesheet.get('FauxBold') else font_props_default['weight'],
                        style='italic' if stylesheet.get('FauxItalic') else font_props_default['style'],
                    )
                    if data_list and (data_list[-1]['font'], color) == (font, color):
                        data_list[-1]['text'] += sub_text
                    else:
                        data_list.append(dict(
                            text=sub_text,
                            font=font,
                            color=color
                        ))
                    index += length
                nodes[name] = cls.text_node_cls(rect, data_list)
            else:
                nodes[name] = cls.image_node_cls(rect, layer.composite())
        return cls(psd.width, psd.height, nodes)


def get_font_color(stylesheet):
    if 'FillColor' in stylesheet:
        values = stylesheet['FillColor']['Values']
    else:
        values = stylesheet['StrokeColor']['Values']
    values = [int(x * 255) for x in values]
    return tuple([*values[1:], values[0]])


def get_psd_names(psd, short, suffix):
    lid_map = {}
    for name, lid in sorted(((slugify(layer.name, True), layer.layer_id) for layer in psd), key=lambda x: x[0]):
        _short = len(name) if short is None else short
        ss = ""
        while True:
            _name = Unicode(name)[:_short] + ss
            if _name not in lid_map:
                break
            _short += 1
            ss += suffix
        lid_map[_name] = lid
    return {v: k for k, v in lid_map.items()}



import os
from io import BytesIO
from PIL import Image
from dektools.file import sure_dir
from dektools.dict import is_list


def resize_image(image: Image, outputs, sizes=None):
    kwargs = {}
    if sizes:
        image = image.resize(max(sizes), Image.Resampling.LANCZOS)
        kwargs['sizes'] = sizes
    if not is_list(outputs):
        outputs = [outputs]
    for output in outputs:
        sure_dir(os.path.dirname(output))
        image.save(output, **kwargs)


def to_jpg(bs):
    image = Image.open(BytesIO(bs))
    image.convert('RGB')
    bi = BytesIO()
    image.save(bi, format='JPEG')
    return bi.getvalue()
